/*
Ported by George Ang<gnap.an<AT>gmail.com>
Copyright (C)2003 Barry Dunne (http://www.emule-project.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 
#include <inttypes.h>
#include <list>
#include <map>
#include <kademlia/defines.hpp>
#include <boost/thread/mutex.hpp>
#include <kademlia/utils/debug.hpp>


#ifndef KBUCKET_HPP
#define KBUCKET_HPP

namespace kademlia
{
	template<typename _uintX> class contact;
	template<typename _uintX, class debugger> class routing_zone;

	template<typename _uintX, class debugger = debug>
	       	class kbucket
		{
			public:

				typedef _uintX uint_type;

				typedef contact<uint_type> contact_type;
				typedef std::list<contact_type*> contact_list;
				typedef typename std::list<contact<_uintX>* >::iterator contact_list_iterator;
				typedef typename std::list<contact<_uintX>* >::const_iterator contact_list_const_iterator;
				typedef std::map<uint_type, contact_type* > contact_map_type;

				friend class routing_zone<uint_type,debugger>;

				~kbucket();

				private:
				kbucket();

				kbucket(contact_list &nodes);

				bool add_contact(contact_type *contact_ptr);

				bool set_alive(contact_type *contact_ptr);

				void remove_contact(contact_type *contact_ptr);

				contact_type* get_contact(const uint_type &node_id);

				contact_type* get_oldest();

				size_t get_size() const;

				size_t get_remaining() const;

				void get_all_nodes(contact_list *list_result, bool empty_first = true);

				uint32_t get_closest_to(uint8_t max_type, 
						uint_type const&target, 
						uint32_t max_required,
						contact_map_type *map_result,
					       	bool empty_first = true,
					       	bool set_in_use = false);
					
				bool m_dont_delete_contacts;
				boost::mutex m_mutex;
				contact_list	m_nodes;
		};

	template<typename _uintX, class debugger>
		kbucket<_uintX, debugger>::~kbucket()
		{
			try
			{
				if (!m_dont_delete_contacts)
				{
					boost::mutex::scoped_lock lock(m_mutex);
					// Delete all contacts
					for( contact_list_const_iterator iter = m_nodes.begin();
							iter != m_nodes.end();
							++iter)
						delete *iter;
				}
				// Remove all contact entries.
				m_nodes.clear();
			}
			catch (...)
			{
				debugger::send(RUNTIME, "kbucket destruction exception!");
			}

		}

	template<typename _uintX, class debugger>
		kbucket<_uintX, debugger>::kbucket()
		{
			//init delete contact flag.
			m_dont_delete_contacts = false;
		}

	template<typename _uintX, class debugger>
		kbucket<_uintX, debugger>::kbucket(contact_list &nodes)
		{
			m_nodes = nodes;
			m_dont_delete_contacts = false;
		}

	template<typename _uintX, class debugger>
		bool kbucket<_uintX, debugger>::add_contact(contact_type* contact_ptr)
		{
			assert(contact_ptr!= NULL);
			// Check if we already have a contact with this ID in the list.
			contact_type *contact_test = get_contact((contact_ptr->get_node_id()));
			if (contact_test == NULL)
			{
				// If not full, add to end of list
				if ( m_nodes.size() < K)
				{
					boost::mutex::scoped_lock lock(m_mutex);
					uint_type distance = contact_ptr->get_node_id();
					distance._xor(kad<uint_type, debugger>::get_kad_id());
					contact_ptr->set_distance(distance);
					m_nodes.push_back(contact_ptr);

					return true;
				}
			}

			return false;

		}
	template<typename _uintX, class debugger>
		bool kbucket<_uintX, debugger>::set_alive(contact_type *contact_ptr)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			// Find contact with Node ID.
			for (contact_list_iterator iter = m_nodes.begin();
					iter != m_nodes.end();
					++iter)
			{
				contact_type* iter_con_ptr = *iter;
				if ( contact_ptr->is_same_contact(iter_con_ptr) )
				{
					// Mark contact as being alive.
					iter_con_ptr->update_type();

					// Move to the end of the list
					m_nodes.remove(iter_con_ptr);

					if ( contact_ptr->get_node_id() == iter_con_ptr->get_node_id() )
					{
						m_nodes.push_back(iter_con_ptr);
						return true;
					}
					else
					{
						debugger::send(ROUTETABLE, "contact %s changed node id to %s.", 
								iter_con_ptr->get_node_id().get_hex_string().c_str(),
								contact_ptr->get_node_id().get_hex_string().c_str());

						return false;
					}
				}
			}
			return false;
		}

	template<typename _uintX, class debugger>
		void kbucket<_uintX, debugger>::remove_contact(contact_type *contact_ptr)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			m_nodes.remove(contact_ptr);
		}

	template<typename _uintX, class debugger>
		typename kbucket<_uintX, debugger>::contact_type*
		kbucket<_uintX, debugger>::get_contact(uint_type const& node_id)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			// Find contact by ID.
			for (contact_list_const_iterator iter = m_nodes.begin();
					iter != m_nodes.end();
					++iter)
			{
				if (node_id == (*iter)->get_node_id())
					return *iter;
			}
			return NULL;

		}

	template<typename _uintX, class debugger>
		typename kbucket<_uintX, debugger>::contact_type* kbucket<_uintX, debugger>::get_oldest()
		{
			// All new/updated entries are appended to the back.
			if (m_nodes.size() > 0)
				return m_nodes.front();
			return NULL;

		}
	template<typename _uintX, class debugger>
		size_t kbucket<_uintX, debugger>::get_size() const
		{
			return m_nodes.size();
		}

	template<typename _uintX, class debugger>
		size_t kbucket<_uintX, debugger>::get_remaining() const
		{
			return K - m_nodes.size();
		}

	template<typename _uintX, class debugger>
		void kbucket<_uintX, debugger>::get_all_nodes(contact_list *list_result, bool empty_first)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			// Clear results if requested first.
			if (empty_first)
				list_result->clear();
			// Append all entries to the results.
			if (m_nodes.size() > 0)
				list_result->insert(
						list_result->end(),
						m_nodes.begin(),
						m_nodes.end());

		}

	template<typename _uintX, class debugger>
		uint32_t kbucket<_uintX, debugger>::get_closest_to(
				uint8_t max_type, 
				uint_type const&target, 
				uint32_t max_required,
				contact_map_type* map_result,
				bool empty_first,
				bool set_in_use)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			// Empty list if requested.
			if (empty_first)
				map_result->clear();

			// Return 0 since we have no entries.
			if (m_nodes.size() == 0)
				return 0;
			// First put results in sort order for uTarget so we can insert them correctly.
			// We don't care about max results at this time.

			for ( contact_list_const_iterator iter = m_nodes.begin();
					iter != m_nodes.end();
					++iter)
			{
				if((*iter)->get_type() <= max_type)
				{
					uint_type target_distance((*iter)->get_node_id());
					target_distance._xor(target);
					(*map_result)[target_distance] = *iter;

					// This list will be used for an unknown time, Inc in use so it's not deleted.
					if(set_in_use)
						(*iter)->inc_use();
				}
			}

			// Remove any extra results by least wanted first.
			while(map_result->size() > max_required)
			{
				// Dec in use count.
				if(set_in_use)
					(--map_result->end())->second->dec_use();

				// remove from results
				map_result->erase(--map_result->end());
			}

			// Return result count to the caller.
			return map_result->size();

		}

}

#endif // KBUCKET_HPP


