/*
Copyright (C)2006 George Ang (gnap.an AT gmail.com)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <boost/thread/mutex.hpp>
#include <kademlia/findnode.hpp>
#include <kademlia/findvalue.hpp>
#include <kademlia/storevalue.hpp>
#include <kademlia/publish.hpp>
#include <kademlia/utils/debug.hpp>

#ifndef SESSIONMANAGER_HPP
#define SESSIONMANAGER_HPP

namespace kademlia
{
	template<typename _uintX, class debugger> class session;
//	template<typename _uintX, class debugger> class search_find_node;
//	template<typename _uintX, class debugger> class store_value;
	template<typename _uintX, class debugger> class session_publish;

	template<typename _uintX, class debugger = debug>
	class session_manager
	{
			typedef _uintX uint_type;
			typedef contact<uint_type> contact_type;
			typedef session<uint_type, debugger> session_type;
			typedef std::map<uint_type, session_type*> session_map;
		public:
			typedef typename std::map<_uintX, session_type*>::iterator session_map_iterator;
		public:

			session_manager();
			
			static bool start_session(session_type* session_ptr);

			static void stop_session(session_type* session_ptr, bool delay_delete = false);

			static void stop_all_sessions();

			static void on_timer();

			static bool target_already_exists(uint_type const& session_id);

			static session_type* get_session(uint_type session_id);

			static search_find_node<uint_type,debugger>*
			       	find_node(const uint_type &id, bool complete = false);
			
			static search_find_value<uint_type,debugger>*
       				find_value(std::string const& keyword,
						std::string const& args,
						bool search_first = false);

			static search_find_value<uint_type,debugger>*
       				find_value(uint_type const& target_id,
						std::string const& args,
						bool search_first = false);

			static session_store_value<uint_type,debugger>*
       				store_value(std::string const& keyword,
						std::string const& content,
						std::string const& args,
						bool search_first = false);

			static session_store_value<uint_type,debugger>*
       				store_value(uint_type const& target,
						std::string const& content,
						std::string const& args,
						bool search_first = false);

			static session_publish<uint_type,debugger>*
			       	publish(contact<uint_type> *contact_ptr);

			static uint_type gen_session_id();
	
		private:
			static boost::mutex m_mutex;
			static session_map m_map_sessions;
			static uint_type m_cur_session_id;
	};

	template<typename _uintX, class debugger>
		session_manager<_uintX, debugger>::session_manager()
		{
			m_map_sessions.clear();
		}
			
	template<typename _uintX, class debugger>
		bool session_manager<_uintX, debugger>::start_session(session_type* session_ptr)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			if (target_already_exists(session_ptr->get_session_id()))
			{
				delete session_ptr;
				return false;
			}	
			m_map_sessions[session_ptr->get_session_id()] = session_ptr;

			session_ptr->start();

			debugger::send(SESSION,
					"session on target %s started.",
					session_ptr->get_target_id().get_hex_string().c_str());
			return true;
		}

	template<typename _uintX, class debugger>
		void session_manager<_uintX, debugger>::stop_session(
				session_type* session_ptr,
				bool delay_delete)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			for (session_map_iterator iter = m_map_sessions.begin();
					iter != m_map_sessions.end();
					++iter)
			{
				if(iter->second == session_ptr)
				{
					if(delay_delete)
					{
						debugger::send(SESSION,
								"session on target %s, would be deleted after 15s.",
								session_ptr->get_target_id().get_hex_string().c_str());
						iter->second->prepare_to_stop();
					}
					else
					{
						debugger::send(SESSION,
								"session on target %s, deleted .",
								session_ptr->get_target_id().get_hex_string().c_str());
						delete  iter->second;
						m_map_sessions.erase(iter);// iter = xxxxxx
					}
					return;
				}
			}

		}

	template<typename _uintX, class debugger>
		void session_manager<_uintX, debugger>::stop_all_sessions()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			session_type* session_ptr = NULL;
			for ( session_map_iterator iter = m_map_sessions.begin();
					iter != m_map_sessions.end();
					++iter)
			{
				session_ptr = iter->second;
				debugger::send(SESSION,
						"session on target %s, deleted .",
						session_ptr->get_target_id().get_hex_string().c_str());
				delete iter->second;
			}
			m_map_sessions.clear();

		}

	template<typename _uintX, class debugger>
		bool session_manager<_uintX, debugger>::target_already_exists(uint_type const& session_id)
		{
			return (m_map_sessions.count(session_id) > 0);
		}

	template<typename _uintX, class debugger>
		void session_manager<_uintX, debugger>::on_timer()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			session_type* session_ptr = NULL;
			for (session_map_iterator iter = m_map_sessions.begin();
					iter != m_map_sessions.end();
					++iter)
			{
				session_ptr= iter->second;
				if(session_ptr->timed_out())
				{
					debugger::send(TIMER,
							"session on target %s deleted .",
							session_ptr->get_target_id().get_hex_string().c_str());
					delete session_ptr;
					m_map_sessions.erase(iter); // iter = xxxxxx
				}
				else if (session_ptr->is_finished())
				{

					debugger::send(TIMER,
							"session on target %s, prepared to stop.",
							session_ptr->get_target_id().get_hex_string().c_str());

					session_ptr->prepare_to_stop();
				}
				else
					session_ptr->on_timer();
			}


		}

	template<typename _uintX, class debugger>
		class session_manager<_uintX, debugger>::session_type* 
		session_manager<_uintX, debugger>::get_session(uint_type session_id)
		{
			if (target_already_exists(session_id))
				return m_map_sessions[session_id];
			return NULL;
		}

	template<typename _uintX, class debugger>
		class search_find_node<_uintX, debugger>*
		session_manager<_uintX, debugger>::find_node(uint_type const& id, bool complete)
		{
			search_find_node<uint_type, debugger>* session_ptr = new search_find_node<uint_type, debugger>(id, complete);
			start_session(session_ptr);
			return session_ptr;
		}
			
	template<typename _uintX, class debugger>
		class search_find_value<_uintX, debugger>*
		session_manager<_uintX, debugger>::find_value(
				std::string const& keyword,
				std::string const& args,
				bool search_first) 
		{
			search_find_value<uint_type, debugger>* session_ptr = new search_find_value<uint_type, debugger>(keyword, args, search_first);
			start_session(session_ptr);
			return session_ptr;
		}

	template<typename _uintX, class debugger>
		class search_find_value<_uintX, debugger>*
		session_manager<_uintX, debugger>::find_value(
				uint_type const& target_id,
				std::string const& args,
				bool search_first) 
		{
			search_find_value<uint_type, debugger>* session_ptr = new search_find_value<uint_type, debugger>(target_id, args, search_first);
			start_session(session_ptr);
			return session_ptr;
		}

	template<typename _uintX, class debugger>
		class session_store_value<_uintX, debugger>*
		session_manager<_uintX, debugger>::store_value(
				std::string const& keyword,
				std::string const& content,
				std::string const& args,
				bool search_first) 
		{
			session_store_value<uint_type, debugger>* session_ptr = new session_store_value<uint_type, debugger>(keyword, content, args, search_first);
			start_session(session_ptr);
			return session_ptr;
		}

	template<typename _uintX, class debugger>
		class session_store_value<_uintX, debugger>*
		session_manager<_uintX, debugger>::store_value(
				uint_type const& target,
				std::string const& content,
				std::string const& args,
				bool search_first) 
		{
			session_store_value<uint_type, debugger>* session_ptr = new session_store_value<uint_type, debugger>(target, content, args, search_first);
			start_session(session_ptr);
			return session_ptr;
		}

	template<typename _uintX, class debugger>
		class session_publish<_uintX, debugger>*
		session_manager<_uintX, debugger>::publish(contact_type* contact_ptr) 
		{
			session_publish<uint_type, debugger>* session_ptr = new session_publish<uint_type, debugger>(contact_ptr);
			start_session(session_ptr);
			return session_ptr;
		}

	template<typename _uintX, class debugger>
		typename session_manager<_uintX, debugger>::uint_type
		session_manager<_uintX, debugger>::gen_session_id()
		{
			m_cur_session_id += 1;
			return m_cur_session_id;
		}

	template<typename _uintX, class debugger>
		typename session_manager<_uintX, debugger>::session_map
		session_manager<_uintX, debugger>::m_map_sessions;

	template<typename _uintX, class debugger>
		boost::mutex session_manager<_uintX, debugger>::m_mutex;

	template<typename _uintX, class debugger>
		typename session_manager<_uintX, debugger>::uint_type
		session_manager<_uintX, debugger>::m_cur_session_id = uint_type(uint_type(false), 0);
};

#endif //SESSION_MANAGER_HPP
